/*
 * main.c
 *
 * main loop, and interface with user
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97 Akim Demaille, Miguel Santana
 */

/*
 * This file is part of a2ps.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * $Id: main.c,v 1.140 1997/11/07 13:28:09 demaille Exp $
 */


/************************************************************************/
/*									*/
/*			I n c l u d e   f i l e s			*/
/*                                                                      */
/************************************************************************/

#include "a2ps.h"
#include "routines.h"
#include "jobs.h"
#include "useropt.h"
#include "confg.h"
#include "options.h"
#include "pathwalk.h"
#include "darray.h"
#include "select.h"
#include "generate.h"
#include "printers.h"
#include "msg.h"
#include "delegate.h"
#include "metaseq.h"
#include "regex.h"
#include "buffer.h"
#include "psgen.h"
#include "prolog.h"
#include "stream.h"

/************************************************************************/
/*									*/
/*		   G l o b a l   d e f i n i t i o n s			*/
/*									*/
/************************************************************************/
/************************************************************************
 *	The various global behaviors
 */
enum behavior {
  b_ps,  	/* postscript generator (usual PS converter) 	*/
  b_guess, 	/* do as file(1) does: return the ssh file name */
  b_help, 
  b_version, 
  b_list_options, 
  b_list_features,
  b_list_media,
  b_list_style_sheets,
  b_list_html_style_sheets,
  b_list_texinfo_style_sheets,
  b_list_printers,
  b_list_delegations,
  b_list_macro_meta_sequences,
  b_list_encodings,
  b_list_texinfo_encodings,
  b_list_user_options
};

enum behavior behavior = b_ps;	/* PS generation			*/

char * program_name;		/* Name of this program 		*/
a2ps_job * job = NULL;		/* The whole information		*/

char * re_syntax_table = NULL;		/* For regex			*/

#define RE_SYNTAX_A2PS (RE_SYNTAX_EMACS | RE_CHAR_CLASSES | RE_INTERVALS)
/************************************************************************/
/*	Related to the config files					*/
/************************************************************************/
/* 
 * Hash table of the delegations
 */
struct hash_table * delegation_table;

/*
 * Content of sheets.map
 */
struct darray * sheets_map = NULL;	/* Content of sheets.map	*/

/*
 * Hash table of the sheet yet read
 */
struct hash_table * style_sheets = NULL;

/************************************************************************/
/*	Related to the options						*/
/************************************************************************/
/*
 * Delegate files to other applications.
 */
int delegate = TRUE;

/*
 * --toc[=format], generate a table of content
 */
uchar * toc = NULL;

/*
 * -E: style sheet to use. NULL => automated
 */
char * style_request = NULL;

/*
 * -g/--highlight-level: 0, 1 or 2
 */
int highlight_level = 1;

/*
 * --strip=NUM, don't write the comments
 */
int strip_level = 0;

/*
 * --end-of-line=TYPE, specify what are the sequences of chars to
 * interpret as end of line
 */
enum eol end_of_line = eol_auto;

/************************************************************************/
/*			       Service routines				*/
/************************************************************************/
/*
 * Unlink all the used files.  Used for atexit
 */
static void
unlink_tmpfiles (void)
{
  if (job)
    a2ps_job_unlink_tmpfiles (job);
}

/*
 * Read the highlighting level
 */
static const char *const highlight_level_args[] =
{
  "none", "off", "0",
  "normal", "light", "1",
  "heavy", "symbols", "2",
  0
};

static const int highlight_level_types[] =
{
  0, 0, 0,
  1, 1, 1,
  2, 2, 2
};

/*
 * Return the highlight_level value
 */
static int
get_highlight_level (const char * option, const char * arg)
{
  int i = get_symbolic_value (highlight_level_args,
			      highlight_level_types,
			      option, arg);
  return highlight_level_types[i];
}

static char *
highlight_level_to_string (int level)
{
  switch (level) {
  case 2:
    /* TRANS: highlighting level = heavy (2/2)*/
    return _("heavy");

  case 1:
    /* TRANS: highlighting level = normal (1/2)*/
    return _("normal");

  case 0:
    /* TRANS: highlighting level = none (0/2) */
    return _("none");
  }
  return NULL; /* For -Wall */
}

/************************************************************************/
/*				Interface				*/
/************************************************************************/
static void
version (void)
{
  printf ("%s %s -- Akim Demaille, Miguel Santana\n", PACKAGE, VERSION);
  printf (_("Send bugs reports and comments to demaille@inf.enst.fr\n"));
}

/*
 * Print information depending on the installation.
 * It is also used to store useful information in the output 
 * to help us debuging the users :).
 */
static void
list_options (struct a2ps_job * a_job, FILE * stream)
{
  uchar buf[256], buf2[256];
  char * cp = NULL;

  /* Title of --list-options (%s%s is `a2ps' `version' */
  fprintf (stream, _("Configuration status of %s %s\n"),
	   program_name, VERSION);

  fputs (_("Sheets:\n"), stream);
  fprintf (stream, _("\
  default paper         = %s\n\
  layout per page       = %d x %d (%s, %s)\n\
  borders               = %s\n\
  compact mode          = %s\n\
  inside margin         = %d\n"), 
	   a_job->medium->name,
	   a_job->columns, a_job->rows, 
	   (a_job->orientation == portrait) ? _("portrait") : _("landscape"),
	   a_job->Major == major_rows ? _("rows first") : _("columns first"),
	   a_job->border ? _("yes") : _("no"),
	   a_job->compact_mode ? _("yes") : _("no"),
	   a_job->margin);

  if (a_job->columns_requested > 0)
    sprintf ((char *) buf, _("%d characters per line"), 
	     a_job->columns_requested);
  else if (a_job->lines_requested > 0)
    sprintf ((char *) buf, _("%d lines per page"), 
	     a_job->lines_requested);
  else
    /* There are no rules to determine the font size.
     * fontsize is "free" */
    sprintf ((char *) buf, _("free"));

  switch (a_job->numbering) {
  case 0:
    ustrcpy (buf2, _("no"));
    break;
  case 1:
    /* number line: each line */
    ustrcpy (buf2, _("each line"));
    break;
  default:
    /* number line: each %d line */
    sprintf ((char *) buf2, _("each %d lines"), a_job->numbering);
  }

  fputs (_("Virtual pages:\n"), stream);
  fprintf (stream, _("\
  number lines          = %s\n\
  format                = %s\n\
  tabulation size       = %d\n\
  non printable format  = %s\n"),
	   buf2,
	   buf,
	   a_job->tabsize,
	   unprintable_format_to_string (a_job->unprintable_format));

  fputs (_("Headers:\n"), stream);
  fprintf (stream, _("\
  header              = %s\n\
  left footer         = %s\n\
  footer              = %s\n\
  right footer        = %s\n\
  left title          = %s\n\
  center title        = %s\n\
  right title         = %s\n\
  under lay           = %s\n"),
	   UNNULL (a_job->header),
	   UNNULL (a_job->left_footer),
	   UNNULL (a_job->footer),
	   UNNULL (a_job->right_footer),
	   UNNULL (a_job->left_title),
	   UNNULL (a_job->center_title),
	   UNNULL (a_job->right_title),
	   UNNULL (a_job->water));

  fputs (_("Input:\n"), stream);
  fprintf (stream, _("\
  truncate lines        = %s\n\
  interpret             = %s\n\
  end of line           = %s\n\
  default encoding      = %s\n\
  document title        = %s\n\
  default prologue      = %s\n\
  print anyway          = %s\n\
  delegating            = %s\n"),
	   a_job->folding ? _("no") : _("yes"),
	   a_job->interpret ? _("yes") : _("no"),
	   eol_to_string (end_of_line),
	   encoding_entry_name (a_job->requested_encoding),
	   a_job->title,
	   a_job->prolog,
	   a_job->print_binaries ? _("yes") : _("no"),
	   delegate ? _("yes") : _("no"));

  /*
   * Pretty printing
   */
  if (IS_EMPTY (style_request))
    /* a2ps -E --list-options */
    ustrcpy (buf, _("selected automatically"));
  else
    ustrcpy (buf, style_request);
  fputs (_("Pretty-printing:\n"), stream);
  fprintf (stream, _("\
  style sheet           = %s\n\
  highlight level       = %s\n\
  strip level           = %d\n"),
	   buf,
	   highlight_level_to_string (highlight_level),
	   strip_level);

  /*
   * Information on where will go the output
   */

  /* Make a nice message to tell where the output is sent */
  destination_to_string (a_job, buf);
  
  /* Make a nice message to tell what version control is used */
  switch (backup_type) {
  case none:
    cp = _("never make backups");
    break;

  case simple:
    cp = _("simple backups of every file");
    break;

  case numbered_existing:
    /* appears in a2ps --version-=existing --list-options */
    cp = _("numbered backups of files already numbered,\n\
                            and simple of others");
    break;

  case numbered:
    cp = _("numbered backups of every file");
    break;
  }

  fputs (_("Output:\n"), stream);
  fprintf (stream, _("\
  destination             = %s\n\
  version control         = %s\n\
  backup suffix           = %s\n"),
	   buf, cp, simple_backup_suffix);
  

  /*
   * PostScript report.
   * TRANS: to be aligned with `statusdict definitions ='
   * and with `page prefeed ='
   */  
  fputs (_("PostScript:\n"), stream);
  fprintf (stream, _("\
  magic number            = %s\n\
  page label format       = %s\n\
  number of copies        = %d\n\
  sides per sheet         = %d\n\
  page device definitions = "),
	   a_job->status->magic_number,
	   a_job->status->page_label_format,
	   a_job->copies,
	   a_job->rectoverso ? 2 : 1);

  list_pagedevice (a_job, stream);
  fprintf (stream, _("\
  statusdict definitions  = "));
  list_statusdict (a_job, stream);
  fprintf (stream, _("\
  page prefeed            = %s\n"),
	   a_job->page_prefeed ? _("yes") : _("no"));

  
  /*
   * Internal Details
   */
  fputs (_("Internals:\n"), stream);
  fprintf (stream, _("\
  verbosity level     = %d\n\
  file command        = %s\n\
  temporary directory = %s\n"),
	   message_verbosity,
	   UNNULL (job->file_command),
	   UNNULL (job->tmpdir));
  fputs (_("\
  library path        = \n"), stream);
  pw_fprintf_path (stream, "\t%s\n", a_job->lib_path);
} 

/*
 * This is used in psgen to push into the PostScript
 * the report of the state of a2ps when it produced the file.
 * I'm tired of fighting with users who don't really say everything
 * on the state of their a2ps when something goes wrong.
 * Now I can ask a ps file generated with --debug, and
 * got everything I need.
 */
static void
list_options_with_prefix (struct a2ps_job * a_job, FILE * stream)
{
#define PREFIX "% "
  FILE * tmp_file;
  char * tmp_filename;
  char buf [BUFSIZ];

  tmp_filename = tmpnam (NULL);
  tmp_file = fopen (tmp_filename, "w");
  if (!tmp_file)
    error (1, errno, _("cannot open file `%s'"), tmp_filename);

  list_options (a_job, tmp_file);
  fclose (tmp_file);

  fopen (tmp_filename, "r");
  if (!tmp_file)
    error (1, errno, _("cannot open file `%s'"), tmp_filename);
  
  while (fgets (buf, sizeof (buf), tmp_file)) {
    fputs (PREFIX, stream);
    fputs (buf, stream);
  }
  fputs (PREFIX, stream);
  putc ('\n', stream);
  fclose (tmp_file);
  unlink (tmp_filename);
}

/*
 * Print information depending on the installation.
 */
static void
list_features (struct a2ps_job * a_job, FILE * stream)
{
  /* Known languages */
  list_style_sheets_short (a_job, stream);
  putc ('\n', stream);

  /* Known char sets */
  list_encodings_short (a_job, stream);
  putc ('\n', stream);

  /* Known media */
  list_media_short (a_job, stream);
  putc ('\n', stream);

  /* Known prologues */
  prologues_list_short (a_job, stream);
  putc ('\n', stream);

  /* Known "printers" */
  list_printers_short (a_job);
  putc ('\n', stream);

  /* Known "delegates" */
  delegations_list_short (delegation_table, stream);
  putc ('\n', stream);

  /* Known user options */
  user_options_list_short (a_job, stream);
  putc ('\n', stream);

  /* Macro meta seq. */
  macro_meta_sequences_list_short (a_job, stream);
  putc ('\n', stream);
}

/*
 * Print a usage message.
 */
static void
usage (void)
{
  printf (_("\
Usage: %s [OPTIONS]... FILES...\n\
   or: %s [OPTIONS]...\n\
To print FILES, or to print the standard input.\n\
Mandatory arguments to long options are mandatory for short options too.\n\
Long options marked with * require a yes/no argument, corresponding\n\
short options stand for `yes'.\n"),
	  program_name, program_name);

  fputs (_("Global behavior:\n"), stdout);
  fputs (_("\
  -q, --quiet, --silent      be really quiet\n\
  -v, --verbose              tell what we are doing\n\
  -V, --version              print version and exit\n\
  -=, --user-option=OPTION   use the user defined shortcut OPTION\n\
  --debug                    enable debugging features\n\
  --guess                    write guessed languages (don't print)\n\
  --help                     print this help and exit\n\
  --list-options             list all options and their values\n\
  --list-features            list media, encodings, languages, prologues,\n\
                             printers, delegates, and user shortcuts\n\
  --list-delegations         detailed list of delegate applications\n\
  --list-encodings           detailed list of encodings\n\
  --list-macro-meta-sequences detailed list of the user's meta sequences\n\
  --list-media               detailed list of media\n\
  --list-printers            detailed list of printers (outputs)\n\
  --list-style-sheets        detailed list of style sheets\n\
  --list-user-options        detailed list of option short cuts\n"),
	 stdout);

  fputs (_("Sheets:\n"), stdout);
  fputs (_("\
  -M, --medium=NAME          use output medium NAME\n\
  -1, -2, ..., -9            predefined layouts for 1, ..., 9 virtual pages\n\
  --columns=NUM              number of columns per sheet (NUM > 0)\n\
  --rows=NUM                 number of rows per sheet (NUM > 0)\n\
  --major=DIRECTION          first fill (DIRECTION=)rows, or columns\n\
  -r, --landscape            print in landscape mode\n\
  -R, --portrait             print in portrait mode\n\
  -j, --borders*             print borders around columns\n\
  -A, --compact*             allow two files on the same sheet\n\
  --margin[=NUM]             define an interior margin of size NUM\n"),
	 stdout);

  fputs (_("Virtual pages:\n"), stdout);
  fputs (_("\
  --line-numbers=NUM         precede each NUM lines with its line number\n\
  -C                         alias for --line-numbers=5\n\
  -f, --font-size=SIZE       use font SIZE (float) for the body text\n\
  -L, --lines-per-page=NUM   specify how many lines are printed on each page\n\
  -l, --chars-per-line=NUM   specify how many columns are printed per page\n\
  -m  --catman               process the file as a man (same as -L66)\n\
  -T, --tabsize=NUM          set tabulator size to NUM\n\
  --non-printable-format=FMT specify how non-printable chars are printed\n"),
	 stdout);

  fputs (_("Headers:\n"), stdout);
  /* xgettext:no-c-format */
  fputs (_("\
  -B, --no-header            no page headers at all\n\
  -b, --header[=TEXT]        set page header\n\
  -u, --underlay[=TEXT]      print TEXT under every page\n\
  --center-title[=TEXT]      set page title to TITLE\n\
  --left-title[=TEXT]        set left and right page title to TEXT\n\
  --right-title[=TEXT]\n\
  --left-footer[=TEXT]       set sheet footers to TEXT\n\
  --footer[=TEXT]            TEXT may use special $ and % escapes\n\
  --right-footer[=TEXT]\n"), stdout);

  fputs (_("Input:\n"), stdout);
  fputs (_("\
  -a, --pages=RANGE          select the pages to print\n\
  -c, --truncate-lines*      cut long lines\n\
  -i  --interpret*           interpret tab, bs and ff chars\n\
  --end-of-line=TYPE         specify the eol char (TYPE: r, n, nr, rn, any)\n\
  -X, --encoding=NAME        use input encoding NAME\n\
  -t, --title=NAME           set the name of the job\n\
  --stdin=NAME               set the name of the input file stdin\n\
  --prologue=FILE            include FILE.pro as PostScript prologue\n\
  --print-anyway*            force binary printing\n\
  -Z, --delegate*            delegate files to another application\n"),
	 stdout);

  fputs (_("Pretty-printing:\n"), stdout);
  fputs (_("\
  -E, --pretty-print[=LANG]  enable pretty-printing (set style to LANG)\n\
  --highligh-level=LEVEL     set pretty printing highlight LEVEL\n\
                             LEVEL can be none, normal or heavy\n\
  -g                         alias for --highligh-level=heavy\n\
  --strip-level=NUM          level of comments stripping\n"),
	 stdout);

  fputs (_("Output:\n"), stdout);
  fputs (_("\
  -o, --output=FILE          leave output to file FILE.  If FILE is `-',\n\
                             leave output to stdout.\n\
  --version-control=WORD     override the usual version control\n\
  --suffix=SUFFIX            override the usual backup suffix\n\
  -P, --printer=NAME         send output to printer NAME\n\
  -d                         send output to the default printer\n"),
	 stdout);

  fputs (_("PostScript:\n"), stdout);
  fputs (_("\
  -n, --copies=NUM           print NUM copies of each page\n\
  -s, --sides=NUM            number of sheet sides (recto/recto-verso)\n\
  -D, --setpagedevice=K[:V]  pass a page device definition to output\n\
  -S, --statusdict=K[:[:]V]  pass a statusdict definition to the output\n\
  -k, --page-prefeed         enable page prefeed\n\
  -K, --no-page-prefeed      disable page prefeed\n"),
	 stdout);
}

/*
 * Handle the options that a2ps understands (not liba2ps)
 * Return 1 for success
 */
static int
handle_a2ps_option (int option, const char * optional_arg)
{
  switch (option) {
  case 'E':           		/* --pretty-print select language */
    xstrcpy (style_request, optional_arg);
    break;
    
  case 'g':				/* Symbol translation */
    highlight_level = 2;
    break;
    
  case 154:				/* Symbol translation */
    highlight_level = 
      get_boolean ("--graphic-symbols", optional_arg);
    break;
    
  case 173:				/* Level of high lighting */
    highlight_level = 
      get_highlight_level ("--highlight-level", optional_arg);
    break;
    
  case 'h':				/* --help */
    behavior = b_help;
    break;
    
    
  case 'V':				/* version and configuration info */
    behavior = b_version;
    break;
    
  case 'Z':				/* --delegate */
    delegate = TRUE;
    break;
    
  case 137:				/* --list-printers */
    behavior = b_list_printers;
    break;
    
  case 138:
    behavior = b_guess;
    break;
    
  case 139:
    behavior = b_list_options;
    break;
    
  case 145:
    behavior = b_list_features;
    break;
    
  case 148:				/* --strip-level */
    strip_level =
      get_integer_in_range ("--strip-level", optional_arg, 0, 3);
    break;
    
  case 160:				/* pass through */
    delegate = 
      get_boolean ("--delegate", optional_arg);
    break;
    
  case 161:				/* --list-media	*/
    behavior = b_list_media;
    break;
    
  case 162:				/* --list-style-sheets	*/
    behavior = b_list_style_sheets;
    break;
    
  case 163:				/* --list-html-style-sheets	*/
    behavior = b_list_html_style_sheets;
    break;
    
  case 164:				/* --list-delegations */
    behavior = b_list_delegations;
    break;
    
  case 165:				/* --list-texinfo-style-sheets	*/
    behavior = b_list_texinfo_style_sheets;
    break;
    
  case 167:			/* --toc=[toc format]			*/
    xustrcpy (toc, optional_arg);
    break;
    
  case 168:				/* --list-macro-meta-sequences	*/
    behavior = b_list_macro_meta_sequences;
    break;
    
  case 169:				/* --end-of-line=TYPE		*/
    end_of_line = option_string_to_eol ("--end-of-line", optional_arg);
    break;
    
  case 170:				/* --list-encodings	*/
    behavior = b_list_encodings;
    break;
    
  case 171:				/* --list-texinfo-encodings	*/
    behavior = b_list_texinfo_encodings;
    break;
    
  case 172:				/* --list-user-options	*/
    behavior = b_list_user_options;
    break;
    
  default:
    return 0;
  }
  return 1;
}

/************************************************************************/
/*			Main routine for this program.			*/
/************************************************************************/
int
main (int argc, char *argv[])
{
  int argn;

  /* name under which  and path from which this program has been called */
  program_name = strrchr (argv[0], DIR_SEP);
  if (program_name == NULL)
    program_name = argv[0];
  else {
    *program_name = NUL;
    program_name++;
    argv[0] = program_name;
  }

  /* A verbose mode starting even before that a2ps parses the
   * options: call it with a name starting with `y' */
  if (*program_name == 'y')
    message_verbosity = 10;
  else
    message_verbosity = 2;

  /* Hooks for reading the config files */
  delegation_hook = add_delegation;

  /* Hook when reading the options */
  handle_option_hook = handle_a2ps_option;

  /* Hook after having generated the PostScript comments */
  ps_comment_hook = list_options_with_prefix;

  /* Prepare to receive in the hash table of the delegations */
  delegation_table = delegation_table_new ();

  /* Set the syntax that has be chosen for regexp */
  re_set_syntax (RE_SYNTAX_A2PS);

  /* The spine of a2ps */
  job = a2ps_job_new ();

  /* System's config.    */
  a2_read_sys_config (job);

  /* Personal config.: only when installed, because there
   * may be incompatibilities between config files versions */
  if (! (getenv ("NO_SYSCONF")))
    (void) a2_read_config (job, job->user->home, ".a2ps/a2psrc");
  
  /* Local config. */
  (void) a2_read_config (job, ".", ".a2psrc");

  /* Prepare the sheets map structure and the hash table that
   * receives yet read sheets */
  sheets_map = new_sheets_map ();
  style_sheets = new_style_sheets ();

  /* Process global options */
  argn = handle_option (job, argc, argv);
  
  /* Once a2ps.cfg is read, finish the building of a2ps_job */
  a2ps_job_finalize (job);

  /* Attach the arguments to the JOB */
  job->argv = argv;
  job->argc = argc;

  /* Now that the job is complete, attach the tmp files unlinking
   * routine to the exit */
  atexit (unlink_tmpfiles);

  switch (behavior) {
    /* do we have not to print, but return the type of the files */
  case b_guess:
    for (; argn < argc; argn++)
      printf ("[%s (%s)]\n", argv[argn], get_command (argv[argn]));
    break;
  
  case b_version:
    version ();
    break;

  case b_help:
    usage ();
    break;

  case b_list_options:
    list_options (job, stdout);
    break;

  case b_list_features:
    list_features (job, stdout);
    break;

  case b_list_media:
    list_media_long (job, stdout);
    break;

  case b_list_style_sheets:
    list_style_sheets_long (job, stdout);
    break;

  case b_list_html_style_sheets:
    /* This is done to ease the update of a2ps' web page */
    list_html_style_sheets (job, stdout);
    break;

  case b_list_texinfo_style_sheets:
    /* This is done to ease the update of a2ps' Texinfo doc */
    list_texinfo_style_sheets (job, stdout);
    break;

  case b_list_printers:
    list_printers_long (job);
    break;

  case b_list_delegations:
    delegations_list_long (delegation_table, stdout);
    break;

  case b_list_macro_meta_sequences:
    macro_meta_sequences_list_long (job, stdout);
    break;

  case b_list_encodings:
    list_encodings_long (job, stdout);
    break;

  case b_list_texinfo_encodings:
    list_texinfo_encodings_long (job, stdout);
    break;

  case b_list_user_options:
    user_options_list_long (job, stdout);
    break;

    /*
     * Text to PostScript generator
     */
  case b_ps:
    {
      /* Count the number of jobs done */
      int delegated_jobs = 0, native_jobs = 0;

      a2ps_open_output_session (job);

      if (argn == argc)		/* Print stdin */
	print (UNULL, &native_jobs, &delegated_jobs);
      else				/* Print following files */
	for (; argn < argc; argn++) {
	  /* `-' designates stdin */
	  if (strequ (argv[argn], "-"))
	    print (UNULL, &native_jobs, &delegated_jobs);
	  else
	    print ((uchar *) argv[argn], &native_jobs, &delegated_jobs);
	}
      
      if (!IS_EMPTY (toc))
	print_toc ((uchar *) _("Table of Content"), toc, &native_jobs);

      if ((native_jobs == 0) && (delegated_jobs == 1)) {
	/* a2ps has only been used to delegate a single job.
	 * Hence its prologue is superfluous */
	/* FIXME: if there were other files but which failed,
	 * then there is _no_reason_ that the file we're interested
	 * in is this one!
	 * To this end, we need to put more information is the file_job
	 * on how its processing went. */
	a2ps_open_output_stream (job);
	dump_file_content_and_unlink (job->output_stream->fp,
				      CURRENT_FILE (job)->tmp_filename);
	a2ps_close_output_stream (job);
	message_job_pages_printed (job);
      } else if (native_jobs || delegated_jobs) {
	/* The whole stuff is needed */
	a2ps_close_output_session (job);
	message_job_pages_printed (job);
      } else {
	/* Nothing has been printed.
	 * Don't close the job, so that nothing is sent to the printer,
	 * not even the PS prologue */
	message_nothing_printed ();
      }
    }
    break;
  }

  a2ps_job_free (job);
  job = NULL;

  return (EXIT_SUCCESS);
}
